package org.codehaus.jackson.map;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.codehaus.jackson.*;
import org.codehaus.jackson.annotate.JsonTypeInfo.As;
import org.codehaus.jackson.map.jsontype.TypeIdResolver;

/**
 * Interface for deserializing type information from JSON content, to
 * type-safely deserialize data into correct polymorphic instance
 * (when type inclusion has been enabled for type handled).
 *<p>
 * Separate deserialization methods are needed because serialized
 * form for inclusion mechanism {@link As#PROPERTY}
 * is slighty different if value is not expressed as JSON Object:
 * and as such both type deserializer and serializer need to
 * JSON Object form (array, object or other (== scalar)) being
 * used.
 * 
 * @since 1.5
 * @author tatus
 */
public abstract class TypeDeserializer
{
    /**
     * Set of allowed packages for bean deserialization.
     */
    protected final static Set<String> ALLOW_DESER_PACKAGES;

    static {
        String strlist = System.getProperty("jackson.deserialization.whitelist.packages");
        Set<String> s = new HashSet<String>();
        if (strlist != null) {
            s = new HashSet<String>(Arrays.asList(strlist.split(",")));
        }
        ALLOW_DESER_PACKAGES = Collections.unmodifiableSet(s);
    }

    /*
    /**********************************************************
    /* Introspection
    /**********************************************************
     */

    /**
     * Accessor for type information inclusion method
     * that deserializer uses; indicates how type information
     * is (expected to be) embedded in JSON input.
     */
    public abstract As getTypeInclusion();

    /**
     * Name of property that contains type information, if
     * property-based inclusion is used.
     */
    public abstract String getPropertyName();

    /**
     * Accessor for object that handles conversions between
     * types and matching type ids.
     */
    public abstract TypeIdResolver getTypeIdResolver();

    /**
     * Accessor for "default implementation" type; optionally defined
     * class to use in cases where type id is not
     * accessible for some reason (either missing, or can not be
     * resolved)
     * 
     * @since 1.9
     */
    public abstract Class<?> getDefaultImpl();
    
    /*
    /*********************************************************
    /* Type deserialization methods
    /**********************************************************
     */

    /**
     * Method called to let this type deserializer handle 
     * deserialization of "typed" object, when value itself
     * is serialized as JSON Object (regardless of Java type).
     * Method needs to figure out intended
     * polymorphic type, locate {@link JsonDeserializer} to use, and
     * call it with JSON data to deserializer (which does not contain
     * type information).
     */
    public abstract Object deserializeTypedFromObject(JsonParser jp, DeserializationContext ctxt)
        throws IOException, JsonProcessingException;

    /**
     * Method called to let this type deserializer handle 
     * deserialization of "typed" object, when value itself
     * is serialized as JSON Array (regardless of Java type).
     * Method needs to figure out intended
     * polymorphic type, locate {@link JsonDeserializer} to use, and
     * call it with JSON data to deserializer (which does not contain
     * type information).
     */
    public abstract Object deserializeTypedFromArray(JsonParser jp, DeserializationContext ctxt)
        throws IOException, JsonProcessingException;

    /**
     * Method called to let this type deserializer handle 
     * deserialization of "typed" object, when value itself
     * is serialized as a scalar JSON value (something other
     * than Array or Object), regardless of Java type.
     * Method needs to figure out intended
     * polymorphic type, locate {@link JsonDeserializer} to use, and
     * call it with JSON data to deserializer (which does not contain
     * type information).
     */
    public abstract Object deserializeTypedFromScalar(JsonParser jp, DeserializationContext ctxt)
        throws IOException, JsonProcessingException;

    /**
     * Method called to let this type deserializer handle 
     * deserialization of "typed" object, when value itself
     * may have been serialized using any kind of JSON value
     * (Array, Object, scalar). Should only be called if JSON
     * serialization is polymorphic (not Java type); for example when
     * using JSON node representation, or "untyped" Java object
     * (which may be Map, Collection, wrapper/primitive etc).
     */
    public abstract Object deserializeTypedFromAny(JsonParser jp, DeserializationContext ctxt)
        throws IOException, JsonProcessingException;

    /**
     * <p>BZ-1507389 and CVEs related to polymorphic deserialization</p>
     * <p>Check if given type is whitelisted without loading it's class to prevent possible
     * exploit with {@code static} blocks.</p>
     * @param parser
     * @param typeId
     * @throws JsonMappingException
     */
    protected static void checkLegalTypes(JsonParser parser, String typeId)
            throws JsonMappingException {

        if (typeId == null || "".equals(typeId)) {
            return;
        }

        // https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.3.2
        // decompose arrays
        while (typeId.charAt(0) == '[') {
            typeId = typeId.substring(1);
            if (!typeId.isEmpty() && typeId.charAt(0) == 'L') {
                typeId = typeId.substring(1);
            }
            // we don't have to skip last ';' because of startsWith()
        }
        if (typeId.length() == 1) {
            // base type B C D F I J S Z
            return;
        }

        Iterator<String> iter = ALLOW_DESER_PACKAGES.iterator();

        boolean pass = false;

        while (iter.hasNext()) {
            if (typeId.startsWith(iter.next())) {
                pass = true;
                break;
            }
        }

        if (!pass) {
            throw new JsonMappingException(
                    String.format("Illegal type (%s) to deserialize: prevented for security reasons", typeId));
        }
    }

}
    